package org.bitbucket.bradleysmithllc.etlunit.feature.database.db;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.google.gson.stream.JsonWriter;
import edu.emory.mathcs.backport.java.util.Collections;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.DatabaseConnection;
import org.bitbucket.bradleysmithllc.etlunit.util.StringUtils;

import java.io.IOException;
import java.util.*;

public class DatabaseImpl implements Database
{
	private String name;
	private boolean dirty;

	private final List<Catalog> catalogList = new ArrayList<Catalog>();
	private final Map<String, Catalog> catalogMap = new HashMap<String, Catalog>();

	private final DatabaseConnection databaseConnection;

	public DatabaseImpl(DatabaseConnection dc, String name) {
		databaseConnection = dc;
		this.name = name;
	}

	public DatabaseImpl(DatabaseConnection dc) {
		this(dc, null);
	}

	@Override
	public String getName() {
		return name;
	}

	@Override
	public void touch() {
		dirty = true;
	}

	@Override
	public boolean isClean() {
		return !dirty;
	}

	@Override
	public void clear() {
		touch();
		catalogList.clear();
		catalogMap.clear();
	}

	@Override
	public List<Catalog> getCatalogs() {
		return Collections.unmodifiableList(catalogList);
	}

	@Override
	public Map<String, Catalog> getCatalogMap() {
		return catalogMap;
	}

	@Override
	public Catalog getCatalog(String name) {
		String safeNullableUppercase = StringUtils.getSafeNullableUppercase(name);
		Catalog c = catalogMap.get(safeNullableUppercase);

		if (c == null)
		{
			return addCatalog(new CatalogImpl(this, safeNullableUppercase));
		}

		return c;
	}

	@Override
	public DatabaseConnection getDatabaseConnection() {
		return databaseConnection;
	}

	public Catalog addCatalog(Catalog ca)
	{
		String safeNullableUppercase = StringUtils.getSafeNullableUppercase(ca.getName());
		if (catalogMap.containsKey(safeNullableUppercase))
		{
			throw new IllegalArgumentException("Catalog with name [" + safeNullableUppercase + "] already added.");
		}

		catalogMap.put(safeNullableUppercase, ca);
		catalogList.add(ca);

		return ca;
	}

	@Override
	public String toString() {
		return name;
	}

	@Override
	public void toJson(JsonWriter writer) throws IOException {
		writer.name("name");

		if (name == null)
		{
			writer.nullValue();
		}
		else
		{
			writer.value(name);
		}

		writer.name("catalogs").beginArray();

		for (Catalog table : catalogList)
		{
			writer.beginObject();
			table.toJson(writer);
			writer.endObject();
		}

		writer.endArray();

		dirty = false;
	}

	@Override
	public void fromJson(JsonNode node) {
		JsonNode jsonNode = node.get("name");

		if (!jsonNode.isNull())
		{
			name = jsonNode.asText();
		}

		JsonNode tables = node.get("catalogs");

		if (tables != null && !tables.isNull())
		{
			ArrayNode nodeList = (ArrayNode) tables;

			Iterator<JsonNode> it = nodeList.iterator();

			while (it.hasNext())
			{
				JsonNode tnode = it.next();

				CatalogImpl timpl = new CatalogImpl(this);

				timpl.fromJson(tnode);

				addCatalog(timpl);
			}
		}
	}
}